@@ -35,33 +35,31 @@ UCS_PROFILE_FUNC_VOID(ucp_mem_type_unpack,
     ucp_ep_h ep = worker->mem_type_ep[mem_type];
     ucp_lane_index_t lane;
     unsigned md_index;
-    ucp_mem_h memh;
     ucs_status_t status;
-    uct_rkey_bundle_t rkey_bundle;
+    ucp_mtype_pack_context_t pack_context;
 
     if (recv_length == 0) {
         return;
     }
 
-    memh     = ucs_alloca(ucp_memh_size(worker->context));
     lane     = ucp_ep_config(ep)->key.rma_lanes[0];
     md_index = ucp_ep_md_index(ep, lane);
 
     status = ucp_mem_type_reg_buffers(worker, buffer, recv_length, mem_type,
-                                      md_index, memh, &rkey_bundle);
+                                      md_index, &pack_context);
     if (status != UCS_OK) {
         ucs_fatal("failed to register buffer with mem type domain %s",
                   ucs_memory_type_names[mem_type]);
     }
 
     status = uct_ep_put_short(ucp_ep_get_lane(ep, lane), recv_data, recv_length,
-                              (uint64_t)buffer, rkey_bundle.rkey);
+                              (uint64_t)buffer, pack_context.rkey_bundle.rkey);
     if (status != UCS_OK) {
         ucs_fatal("mem type unpack failed to uct_ep_put_short() %s",
                   ucs_status_string(status));
     }
 
-    ucp_mem_type_unreg_buffers(worker, md_index, memh, &rkey_bundle);
+    ucp_mem_type_unreg_buffers(worker, &pack_context);
 }
 
 UCS_PROFILE_FUNC_VOID(ucp_mem_type_pack,
@@ -73,32 +71,30 @@ UCS_PROFILE_FUNC_VOID(ucp_mem_type_pack,
     ucp_lane_index_t lane;
     ucp_md_index_t md_index;
     ucs_status_t status;
-    ucp_mem_h memh;
-    uct_rkey_bundle_t rkey_bundle;
+    ucp_mtype_pack_context_t pack_context;
 
     if (length == 0) {
         return;
     }
 
-    memh     = ucs_alloca(ucp_memh_size(worker->context));
     lane     = ucp_ep_config(ep)->key.rma_lanes[0];
     md_index = ucp_ep_md_index(ep, lane);
 
     status = ucp_mem_type_reg_buffers(worker, (void *)src, length, mem_type,
-                                      md_index, memh, &rkey_bundle);
+                                      md_index, &pack_context);
     if (status != UCS_OK) {
         ucs_fatal("failed to register buffer with mem type domain %s",
                   ucs_memory_type_names[mem_type]);
     }
 
     status = uct_ep_get_short(ucp_ep_get_lane(ep, lane), dest, length,
-                              (uint64_t)src, rkey_bundle.rkey);
+                              (uint64_t)src, pack_context.rkey_bundle.rkey);
     if (status != UCS_OK) {
         ucs_fatal("mem type pack failed to uct_ep_get_short() %s",
                   ucs_status_string(status));
     }
 
-    ucp_mem_type_unreg_buffers(worker, md_index, memh, &rkey_bundle);
+    ucp_mem_type_unreg_buffers(worker, &pack_context);
 }
 
 size_t ucp_dt_pack(ucp_worker_h worker, ucp_datatype_t datatype,
