@@ -62,6 +62,13 @@ static inline void *uct_mem_alloc_params_get_address(const uct_mem_alloc_params_
             params->address : NULL;
 }
 
+static int uct_mem_is_host_alloc(ucs_memory_type_t mem_type,
+                                 ucs_sys_device_t sys_dev)
+{
+    return (mem_type == UCS_MEMORY_TYPE_HOST) &&
+           (sys_dev == UCS_SYS_DEVICE_ID_UNKNOWN);
+}
+
 ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
                            unsigned num_methods,
                            const uct_mem_alloc_params_t *params,
@@ -71,6 +78,7 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
     const uct_alloc_method_t *method;
     ucs_log_level_t log_level;
     ucs_memory_type_t mem_type;
+    ucs_sys_device_t sys_dev;
     uct_md_attr_t md_attr;
     ucs_status_t status;
     unsigned flags;
@@ -94,19 +102,22 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
 
     /* set defaults in case some param fields are not set */
     address      = uct_mem_alloc_params_get_address(params);
-    flags        = (params->field_mask & UCT_MEM_ALLOC_PARAM_FIELD_FLAGS) ?
-                   params->flags : (UCT_MD_MEM_ACCESS_LOCAL_READ |
-                                    UCT_MD_MEM_ACCESS_LOCAL_WRITE);
-    alloc_name   = (params->field_mask & UCT_MEM_ALLOC_PARAM_FIELD_NAME) ?
-                   params->name : "anonymous-uct_mem_alloc";
-    mem_type     = (params->field_mask & UCT_MEM_ALLOC_PARAM_FIELD_MEM_TYPE) ?
-                   params->mem_type : UCS_MEMORY_TYPE_HOST;
+    flags        = UCS_PARAM_VALUE(UCT_MEM_ALLOC_PARAM_FIELD, params, flags,
+                                   FLAGS, (UCT_MD_MEM_ACCESS_LOCAL_READ |
+                                           UCT_MD_MEM_ACCESS_LOCAL_WRITE));
+    alloc_name   = UCS_PARAM_VALUE(UCT_MEM_ALLOC_PARAM_FIELD, params, name,
+                                   NAME, "anonymous-uct_mem_alloc");
+    mem_type     = UCS_PARAM_VALUE(UCT_MEM_ALLOC_PARAM_FIELD, params, mem_type,
+                                   MEM_TYPE, UCS_MEMORY_TYPE_HOST);
+    sys_dev      = UCS_PARAM_VALUE(UCT_MEM_ALLOC_PARAM_FIELD, params, sys_device,
+                                   SYS_DEVICE, UCS_SYS_DEVICE_ID_UNKNOWN);
     alloc_length = length;
     log_level    = (flags & UCT_MD_MEM_FLAG_HIDE_ERRORS) ? UCS_LOG_LEVEL_DEBUG :
                    UCS_LOG_LEVEL_ERROR;
 
-    ucs_trace("allocating %s: %s memory length %zu flags 0x%x", alloc_name,
-              ucs_memory_type_names[mem_type], alloc_length, flags);
+    ucs_trace("allocating %s: length %zu %s memory sys_dev %s flags 0x%x",
+              alloc_name, alloc_length, ucs_memory_type_names[mem_type],
+              ucs_topo_sys_device_get_name(sys_dev), flags);
     ucs_log_indent(1);
 
     for (method = methods; method < methods + num_methods; ++method) {
@@ -151,9 +162,8 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
                  * fall-back, because this MD already exposed support for memory
                  * allocation.
                  */
-                status = uct_md_mem_alloc(md, &alloc_length, &address,
-                                          mem_type, flags, alloc_name,
-                                          &memh);
+                status = uct_md_mem_alloc(md, &alloc_length, &address, mem_type,
+                                          sys_dev, flags, alloc_name, &memh);
                 if (status != UCS_OK) {
                     ucs_log(log_level,
                             "failed to allocate %zu bytes using md %s for %s: %s",
@@ -163,9 +173,10 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
                 }
 
                 ucs_assert(memh != UCT_MEM_HANDLE_NULL);
-                mem->md       = md;
-                mem->mem_type = mem_type;
-                mem->memh     = memh;
+                mem->md         = md;
+                mem->mem_type   = mem_type;
+                mem->sys_device = sys_dev;
+                mem->memh       = memh;
                 goto allocated;
             }
 
@@ -181,7 +192,7 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
             break;
 
         case UCT_ALLOC_METHOD_THP:
-            if (mem_type != UCS_MEMORY_TYPE_HOST) {
+            if (!uct_mem_is_host_alloc(mem_type, sys_dev)) {
                 break;
             }
 
@@ -223,7 +234,7 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
             break;
 
         case UCT_ALLOC_METHOD_HEAP:
-            if (mem_type != UCS_MEMORY_TYPE_HOST) {
+            if (!uct_mem_is_host_alloc(mem_type, sys_dev)) {
                 break;
             }
 
@@ -245,7 +256,7 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
             break;
 
         case UCT_ALLOC_METHOD_MMAP:
-            if (mem_type != UCS_MEMORY_TYPE_HOST) {
+            if (!uct_mem_is_host_alloc(mem_type, sys_dev)) {
                 break;
             }
 
@@ -264,7 +275,7 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
             break;
 
         case UCT_ALLOC_METHOD_HUGE:
-            if (mem_type != UCS_MEMORY_TYPE_HOST) {
+            if (!uct_mem_is_host_alloc(mem_type, sys_dev)) {
                 break;
             }
 
@@ -293,7 +304,10 @@ ucs_status_t uct_mem_alloc(size_t length, const uct_alloc_method_t *methods,
         }
     }
 
-    ucs_debug("could not allocate memory with any of the provided methods");
+    ucs_debug("could not allocate %s: %s memory length=%zu flags=0x%x "
+              "num_methods=%u",
+              alloc_name, ucs_memory_type_names[mem_type], alloc_length, flags,
+              num_methods);
     status = UCS_ERR_NO_MEMORY;
     goto out;
 
@@ -355,30 +369,33 @@ ucs_status_t uct_iface_mem_alloc(uct_iface_h tl_iface, size_t length, unsigned f
                                  const char *name, uct_allocated_memory_t *mem)
 {
     static uct_alloc_method_t method_md = UCT_ALLOC_METHOD_MD;
+    const ucs_memory_type_t mem_type    = UCS_MEMORY_TYPE_HOST;
     uct_base_iface_t *iface             = ucs_derived_of(tl_iface,
                                                          uct_base_iface_t);
     void *address                       = NULL;
-    uct_md_attr_t md_attr;
-    ucs_status_t status;
+    int need_mem_reg, support_mem_reg;
+    uct_alloc_method_t *alloc_methods;
     uct_mem_alloc_params_t params;
     unsigned num_alloc_methods;
-    uct_alloc_method_t *alloc_methods;
+    uct_md_attr_t md_attr;
+    ucs_status_t status;
 
     status = uct_md_query(iface->md, &md_attr);
     if (status != UCS_OK) {
         goto err;
     }
 
-    if (!(md_attr.cap.flags & UCT_MD_FLAG_REG) &&
+    need_mem_reg    = !!(md_attr.cap.flags &
+                         (UCT_MD_FLAG_NEED_MEMH | UCT_MD_FLAG_NEED_RKEY));
+    support_mem_reg = (md_attr.cap.flags & UCT_MD_FLAG_REG) &&
+                      (md_attr.cap.reg_mem_types & UCS_BIT(mem_type));
+
+    if (need_mem_reg && !support_mem_reg &&
+        (md_attr.cap.flags & UCT_MD_FLAG_ALLOC) &&
         uct_iface_is_allowed_alloc_method(iface, UCT_ALLOC_METHOD_MD)) {
-        /* If MD does not support registration, allow only the MD method */
+        /* Force allocation using MD */
         alloc_methods     = &method_md;
         num_alloc_methods = 1;
-    } else if (!(md_attr.cap.flags & UCT_MD_FLAG_REG)) {
-        /* If MD does not support registration and MD method is not in list
-         * return error */
-        status = UCS_ERR_NO_MEMORY;
-        goto err;
     } else {
         alloc_methods     = iface->config.alloc_methods;
         num_alloc_methods = iface->config.num_alloc_methods;
@@ -391,7 +408,7 @@ ucs_status_t uct_iface_mem_alloc(uct_iface_h tl_iface, size_t length, unsigned f
                              UCT_MEM_ALLOC_PARAM_FIELD_NAME;
     params.flags           = flags;
     params.name            = name;
-    params.mem_type        = UCS_MEMORY_TYPE_HOST;
+    params.mem_type        = mem_type;
     params.address         = address;
     params.mds.mds         = &iface->md;
     params.mds.count       = 1;
@@ -402,10 +419,9 @@ ucs_status_t uct_iface_mem_alloc(uct_iface_h tl_iface, size_t length, unsigned f
         goto err;
     }
 
-    /* If the memory was not allocated using MD, register it */
+    /* If the memory was not allocated using MD, register it if needed */
     if (mem->method != UCT_ALLOC_METHOD_MD) {
-        if ((md_attr.cap.flags & UCT_MD_FLAG_REG) &&
-            (md_attr.cap.reg_mem_types & UCS_BIT(mem->mem_type))) {
+        if (need_mem_reg && support_mem_reg) {
             status = uct_md_mem_reg(iface->md, mem->address, mem->length, flags,
                                     &mem->memh);
             if (status != UCS_OK) {
