@@ -78,9 +78,16 @@ static ucs_config_field_t uct_cuda_copy_md_config_table[] = {
      ucs_offsetof(uct_cuda_copy_md_config_t, cuda_async_mem_type),
      UCS_CONFIG_TYPE_ENUM(ucs_memory_type_names)},
 
+    {"RETAIN_PRIMARY_CTX", "n",
+     "Retain and use an inactive CUDA primary context for memory allocation",
+     ucs_offsetof(uct_cuda_copy_md_config_t, retain_primary_ctx),
+     UCS_CONFIG_TYPE_BOOL},
+
     {NULL}
 };
 
+static struct {} uct_cuda_dummy_memh;
+
 static int uct_cuda_copy_md_is_dmabuf_supported()
 {
     int dmabuf_supported = 0;
@@ -118,7 +125,8 @@ uct_cuda_copy_md_query(uct_md_h uct_md, uct_md_attr_v2_t *md_attr)
     md_attr->cache_mem_types  = UCS_BIT(UCS_MEMORY_TYPE_CUDA_MANAGED);
     md_attr->alloc_mem_types  = UCS_BIT(UCS_MEMORY_TYPE_CUDA) |
                                 UCS_BIT(UCS_MEMORY_TYPE_CUDA_MANAGED);
-    md_attr->access_mem_types = UCS_BIT(UCS_MEMORY_TYPE_CUDA) |
+    md_attr->access_mem_types = UCS_BIT(UCS_MEMORY_TYPE_HOST) |
+                                UCS_BIT(UCS_MEMORY_TYPE_CUDA) |
                                 UCS_BIT(UCS_MEMORY_TYPE_CUDA_MANAGED);
     md_attr->detect_mem_types = UCS_BIT(UCS_MEMORY_TYPE_CUDA) |
                                 UCS_BIT(UCS_MEMORY_TYPE_CUDA_MANAGED);
@@ -128,30 +136,6 @@ uct_cuda_copy_md_query(uct_md_h uct_md, uct_md_attr_v2_t *md_attr)
     return UCS_OK;
 }
 
-static ucs_status_t
-uct_cuda_copy_mkey_pack(uct_md_h md, uct_mem_h memh, void *address,
-                        size_t length, const uct_md_mkey_pack_params_t *params,
-                        void *mkey_buffer)
-{
-    return UCS_OK;
-}
-
-static ucs_status_t uct_cuda_copy_rkey_unpack(uct_component_t *component,
-                                              const void *rkey_buffer,
-                                              uct_rkey_t *rkey_p,
-                                              void **handle_p)
-{
-    *rkey_p   = 0xdeadbeef;
-    *handle_p = NULL;
-    return UCS_OK;
-}
-
-static ucs_status_t uct_cuda_copy_rkey_release(uct_component_t *component,
-                                               uct_rkey_t rkey, void *handle)
-{
-    return UCS_OK;
-}
-
 UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_copy_mem_reg,
                  (md, address, length, params, memh_p),
                  uct_md_h md, void *address, size_t length,
@@ -165,7 +149,8 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_copy_mem_reg,
 
     if (!uct_cuda_base_is_context_active()) {
         ucs_debug("attempt to register memory without active context");
-        return uct_md_dummy_mem_reg(md, address, length, params, memh_p);
+        *memh_p = &uct_cuda_dummy_memh;
+        return UCS_OK;
     }
 
     result = cuPointerGetAttribute(&memType, CU_POINTER_ATTRIBUTE_MEMORY_TYPE,
@@ -174,7 +159,8 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_copy_mem_reg,
                                      (memType == CU_MEMORYTYPE_UNIFIED) ||
                                      (memType == CU_MEMORYTYPE_DEVICE))) {
         /* only host memory not allocated by cuda needs to be registered */
-        return uct_md_dummy_mem_reg(md, address, length, params, memh_p);
+        *memh_p = &uct_cuda_dummy_memh;
+        return UCS_OK;
     }
 
     log_level = (flags & UCT_MD_MEM_FLAG_HIDE_ERRORS) ? UCS_LOG_LEVEL_DEBUG :
@@ -194,19 +180,11 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_copy_mem_dereg,
                  (md, params),
                  uct_md_h md, const uct_md_mem_dereg_params_t *params)
 {
-    void *address;
-    ucs_status_t status;
-
     UCT_MD_MEM_DEREG_CHECK_PARAMS(params, 0);
 
-    address = (void *)params->memh;
-    if (address == (void*)0xdeadbeef) {
-        return UCS_OK;
-    }
-
-    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemHostUnregister(address));
-    if (status != UCS_OK) {
-        return status;
+    if (params->memh != &uct_cuda_dummy_memh) {
+        UCT_CUDADRV_FUNC(cuMemHostUnregister((void*)params->memh),
+                         UCS_LOG_LEVEL_DIAG);
     }
 
     return UCS_OK;
@@ -215,7 +193,7 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_copy_mem_dereg,
 static ucs_status_t
 uct_cuda_copy_mem_alloc_fabric(uct_cuda_copy_md_t *md,
                                uct_cuda_copy_alloc_handle_t *alloc_handle,
-                               unsigned flags)
+                               CUdevice cu_device, unsigned flags)
 {
 #if HAVE_CUDA_FABRIC
     CUmemAllocationProp prop    = {};
@@ -223,7 +201,7 @@ uct_cuda_copy_mem_alloc_fabric(uct_cuda_copy_md_t *md,
     ucs_log_level_t log_level   = (md->config.enable_fabric == UCS_YES) ?
                                   UCS_LOG_LEVEL_ERROR : UCS_LOG_LEVEL_DEBUG;
     ucs_status_t status;
-    CUdevice cu_device;
+    int allowed_types;
 
     if (!(flags & UCT_MD_MEM_FLAG_HIDE_ERRORS) &&
         (md->config.enable_fabric == UCS_YES)) {
@@ -232,15 +210,11 @@ uct_cuda_copy_mem_alloc_fabric(uct_cuda_copy_md_t *md,
         log_level = UCS_LOG_LEVEL_DEBUG;
     }
 
-    status = UCT_CUDADRV_FUNC(cuCtxGetDevice(&cu_device), log_level);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    prop.type                 = CU_MEM_ALLOCATION_TYPE_PINNED;
-    prop.requestedHandleTypes = CU_MEM_HANDLE_TYPE_FABRIC;
-    prop.location.type        = CU_MEM_LOCATION_TYPE_DEVICE;
-    prop.location.id          = cu_device;
+    prop.type                            = CU_MEM_ALLOCATION_TYPE_PINNED;
+    prop.requestedHandleTypes            = CU_MEM_HANDLE_TYPE_FABRIC;
+    prop.location.type                   = CU_MEM_LOCATION_TYPE_DEVICE;
+    prop.location.id                     = cu_device;
+    prop.allocFlags.gpuDirectRDMACapable = 1;
 
     if (md->granularity == SIZE_MAX) {
         status = UCT_CUDADRV_FUNC(cuMemGetAllocationGranularity(
@@ -286,9 +260,24 @@ uct_cuda_copy_mem_alloc_fabric(uct_cuda_copy_md_t *md,
         goto err_mem_unmap;
     }
 
+    status = UCT_CUDADRV_FUNC(
+            cuPointerGetAttribute(&allowed_types,
+                                  CU_POINTER_ATTRIBUTE_ALLOWED_HANDLE_TYPES,
+                                  alloc_handle->ptr),
+            log_level);
+    if (status != UCS_OK) {
+        goto err_mem_unmap;
+    } else if (!(allowed_types & CU_MEM_HANDLE_TYPE_FABRIC)) {
+        ucs_log(log_level,
+                "allocated memory at %p of size %ld does not have fabric "
+                "property",
+                (void*)alloc_handle->ptr, alloc_handle->length);
+        goto err_mem_unmap;
+    }
+
     alloc_handle->is_vmm = 1;
 
-    ucs_trace("allocated vmm fabric memory at %p of size %ld\n",
+    ucs_trace("allocated vmm fabric memory at %p of size %ld",
               (void*)alloc_handle->ptr, alloc_handle->length);
     return UCS_OK;
 
@@ -306,20 +295,13 @@ err_mem_release:
 
 typedef CUresult (*uct_cuda_cuCtxSetFlags_t)(unsigned);
 
-static void uct_cuda_copy_sync_memops(uct_cuda_copy_md_t *md,
-                                      const void *address, int is_vmm)
+static ucs_status_t uct_cuda_copy_set_ctx_sync_memops(int log_level)
 {
-    unsigned sync_memops_value = 1;
 #if HAVE_CUDA_FABRIC
     static uct_cuda_cuCtxSetFlags_t cuda_cuCtxSetFlags_func =
         (uct_cuda_cuCtxSetFlags_t)ucs_empty_function;
     CUdriverProcAddressQueryResult sym_status;
     CUresult cu_err;
-    ucs_status_t status;
-
-    if (md->sync_memops_set) {
-        return;
-    }
 
     if (cuda_cuCtxSetFlags_func ==
         (uct_cuda_cuCtxSetFlags_t)ucs_empty_function) {
@@ -334,38 +316,155 @@ static void uct_cuda_copy_sync_memops(uct_cuda_copy_md_t *md,
 
     if (cuda_cuCtxSetFlags_func != NULL) {
         /* Synchronize future DMA operations for all memory types */
-        status = UCT_CUDADRV_FUNC_LOG_WARN(
-                    cuda_cuCtxSetFlags_func(CU_CTX_SYNC_MEMOPS));
-        if (status == UCS_OK) {
-            md->sync_memops_set = 1;
-        }
-
-        return;
+        UCT_CUDADRV_FUNC(cuda_cuCtxSetFlags_func(CU_CTX_SYNC_MEMOPS),
+                         log_level);
+        return UCS_OK;
     }
 #endif
 
-    if (is_vmm) {
+    return UCS_ERR_UNSUPPORTED;
+}
+
+static void uct_cuda_copy_sync_memops(CUdeviceptr dptr, int is_vmm)
+{
+    unsigned sync_memops_value = 1;
+
+    if (uct_cuda_copy_set_ctx_sync_memops(UCS_LOG_LEVEL_WARN) == UCS_OK) {
+        return;
+    } else if (is_vmm) {
         ucs_warn("cannot set sync_memops on CUDA VMM without cuCtxSetFlags() "
-                 "(address=%p)", address);
+                 "(address=0x%llx)",
+                 dptr);
         return;
     }
 
     /* Synchronize for DMA for legacy memory types */
     UCT_CUDADRV_FUNC_LOG_WARN(
             cuPointerSetAttribute(&sync_memops_value,
-                                  CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,
-                                  (CUdeviceptr)address));
+                                  CU_POINTER_ATTRIBUTE_SYNC_MEMOPS, dptr));
+}
+
+ucs_status_t uct_cuda_copy_push_ctx(CUdevice device, int retain_inactive,
+                                    ucs_log_level_t log_level)
+{
+    ucs_status_t status;
+    CUcontext primary_ctx;
+
+    status = uct_cuda_primary_ctx_retain(device, retain_inactive, &primary_ctx);
+    if (status != UCS_OK) {
+        return status;
+    }
+
+    status = UCT_CUDADRV_FUNC(cuCtxPushCurrent(primary_ctx), log_level);
+    if (status != UCS_OK) {
+        (void)UCT_CUDADRV_FUNC(cuDevicePrimaryCtxRelease(device), log_level);
+    }
+
+    return status;
+}
+
+/*
+ * With a valid sys_dev, the function pushes on the current thread the
+ * corresponding CUDA context.
+ * When sys_dev was specified as unknown, the function leaves the current CUDA
+ * context untouched. If no context is set, it tries to push the first
+ * available context among all CUDA GPUs found.
+ */
+static ucs_status_t
+uct_cuda_copy_push_alloc_ctx(uct_cuda_copy_md_t *md,
+                             const ucs_sys_device_t sys_dev,
+                             CUdevice *cu_device_p,
+                             CUdevice *alloc_cu_device_p,
+                             ucs_log_level_t log_level)
+{
+    ucs_status_t status;
+    int dev_ordinal, num_devices;
+
+    status = UCT_CUDADRV_FUNC_LOG_DEBUG(cuCtxGetDevice(cu_device_p));
+    if (status != UCS_OK) {
+        *cu_device_p = CU_DEVICE_INVALID;
+    }
+
+    if (sys_dev != UCS_SYS_DEVICE_ID_UNKNOWN) {
+        status = uct_cuda_base_get_cuda_device(sys_dev, alloc_cu_device_p);
+        if (status != UCS_OK) {
+            ucs_log(log_level,
+                    "failed to get cuda device for system device %u",
+                    sys_dev);
+            return UCS_ERR_INVALID_PARAM;
+        }
+
+        /* sys_dev is the active cuda device */
+        if (*cu_device_p == *alloc_cu_device_p) {
+            return UCS_OK;
+        }
+
+        /* Make sys_dev the active cuda device */
+        status = uct_cuda_copy_push_ctx(*alloc_cu_device_p,
+                                        md->config.retain_primary_ctx,
+                                        log_level);
+        if (status != UCS_OK) {
+            ucs_log(log_level, "failed to set cuda context for system device %u "
+                    "(cu_device=%d)",
+                    sys_dev, *alloc_cu_device_p);
+        }
+
+        return status;
+    }
+
+    /* Use the active cuda device */
+    if (*cu_device_p != CU_DEVICE_INVALID) {
+        *alloc_cu_device_p = *cu_device_p;
+        return UCS_OK;
+    }
+
+    status = UCT_CUDADRV_FUNC(cuDeviceGetCount(&num_devices),
+                              UCS_LOG_LEVEL_DIAG);
+    if (status != UCS_OK) {
+        return UCS_ERR_INVALID_PARAM;
+    }
+
+    /* Use the first active cuda device for allocation */
+    for (dev_ordinal = 0; dev_ordinal < num_devices; dev_ordinal++) {
+        if (UCT_CUDADRV_FUNC_LOG_DEBUG(cuDeviceGet(alloc_cu_device_p, dev_ordinal)) !=
+            UCS_OK) {
+            continue;
+        }
+
+        status = uct_cuda_copy_push_ctx(*alloc_cu_device_p,
+                                        md->config.retain_primary_ctx,
+                                        log_level);
+        if (status == UCS_OK) {
+            break;
+        }
+    }
+
+    if (status != UCS_OK) {
+        ucs_log(log_level,
+                "no active cuda primary context for memory allocation");
+    }
+
+    return status;
+}
+
+static void uct_cuda_copy_pop_alloc_ctx(CUdevice cu_device)
+{
+    (void)UCT_CUDADRV_FUNC(cuCtxPopCurrent(NULL), UCS_LOG_LEVEL_WARN);
+    (void)UCT_CUDADRV_FUNC(cuDevicePrimaryCtxRelease(cu_device),
+                           UCS_LOG_LEVEL_WARN);
 }
 
 static ucs_status_t
 uct_cuda_copy_mem_alloc(uct_md_h uct_md, size_t *length_p, void **address_p,
-                        ucs_memory_type_t mem_type, unsigned flags,
-                        const char *alloc_name, uct_mem_h *memh_p)
+                        ucs_memory_type_t mem_type, ucs_sys_device_t sys_dev,
+                        unsigned flags, const char *alloc_name,
+                        uct_mem_h *memh_p)
 {
     uct_cuda_copy_md_t *md = ucs_derived_of(uct_md, uct_cuda_copy_md_t);
     ucs_status_t status;
     uct_cuda_copy_alloc_handle_t *alloc_handle;
     ucs_log_level_t log_level;
+    CUdevice alloc_cu_device, cu_device;
 
     if ((mem_type != UCS_MEMORY_TYPE_CUDA_MANAGED) &&
         (mem_type != UCS_MEMORY_TYPE_CUDA)) {
@@ -375,12 +474,6 @@ uct_cuda_copy_mem_alloc(uct_md_h uct_md, size_t *length_p, void **address_p,
     log_level = (flags & UCT_MD_MEM_FLAG_HIDE_ERRORS) ? UCS_LOG_LEVEL_DEBUG :
                 UCS_LOG_LEVEL_ERROR;
 
-    if (!uct_cuda_base_is_context_active()) {
-        ucs_log(log_level,
-                "attempt to allocate cuda memory without active context");
-        return UCS_ERR_NO_DEVICE;
-    }
-
     alloc_handle = ucs_malloc(sizeof(*alloc_handle),
                               "uct_cuda_copy_alloc_handle_t");
     if (NULL == alloc_handle) {
@@ -392,9 +485,16 @@ uct_cuda_copy_mem_alloc(uct_md_h uct_md, size_t *length_p, void **address_p,
     alloc_handle->length = *length_p;
     alloc_handle->is_vmm = 0;
 
+    status = uct_cuda_copy_push_alloc_ctx(md, sys_dev, &cu_device,
+                                          &alloc_cu_device, log_level);
+    if (status != UCS_OK) {
+        return UCS_ERR_NO_DEVICE;
+    }
+
     if (mem_type == UCS_MEMORY_TYPE_CUDA) {
         if (md->config.enable_fabric != UCS_NO) {
-            status = uct_cuda_copy_mem_alloc_fabric(md, alloc_handle, flags);
+            status = uct_cuda_copy_mem_alloc_fabric(md, alloc_handle,
+                                                    alloc_cu_device, flags);
             if (status == UCS_OK) {
                 goto allocated;
             } else {
@@ -428,17 +528,22 @@ uct_cuda_copy_mem_alloc(uct_md_h uct_md, size_t *length_p, void **address_p,
 
     if (status != UCS_OK) {
         ucs_free(alloc_handle);
-        return status;
+        goto out;
     }
 
 allocated:
-    uct_cuda_copy_sync_memops(md, (void *)alloc_handle->ptr,
-                              alloc_handle->is_vmm);
+    uct_cuda_copy_sync_memops(alloc_handle->ptr, alloc_handle->is_vmm);
 
     *memh_p    = alloc_handle;
     *address_p = (void*)alloc_handle->ptr;
     *length_p  = alloc_handle->length;
-    return UCS_OK;
+
+out:
+    if (cu_device != alloc_cu_device) {
+        uct_cuda_copy_pop_alloc_ctx(alloc_cu_device);
+    }
+
+    return status;
 }
 
 static ucs_status_t
@@ -522,7 +627,8 @@ static ucs_status_t uct_cuda_copy_mem_free(uct_md_h md, uct_mem_h memh)
     if (alloc_handle->is_vmm) {
         status = uct_cuda_copy_mem_release_fabric(alloc_handle);
     } else {
-        status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemFree(alloc_handle->ptr));
+        UCT_CUDADRV_FUNC(cuMemFree(alloc_handle->ptr), UCS_LOG_LEVEL_DIAG);
+        status = UCS_OK;
     }
 
     ucs_free(alloc_handle);
@@ -571,6 +677,66 @@ err:
     return 1; /* return 1 byte to avoid division by zero */
 }
 
+static void uct_cuda_copy_md_sync_memops_get_address_range(
+        const uct_cuda_copy_md_t *md, CUdeviceptr address, size_t length,
+        CUcontext cuda_ctx, CUdevice cuda_device, int is_vmm,
+        ucs_memory_info_t *mem_info)
+{
+    CUcontext tmp_ctx;
+    CUdeviceptr base_address;
+    size_t alloc_length;
+    size_t total_bytes;
+    ucs_status_t status;
+
+    mem_info->base_address = (void*)address;
+    mem_info->alloc_length = length;
+
+    if (cuda_ctx == NULL) {
+        status = uct_cuda_copy_push_ctx(cuda_device, 0, UCS_LOG_LEVEL_ERROR);
+    } else {
+        status = UCT_CUDADRV_FUNC_LOG_ERR(cuCtxPushCurrent(cuda_ctx));
+    }
+    if (status != UCS_OK) {
+        return;
+    }
+
+    /* Wrapped the method by push/pop CUDA context since it sets
+       CU_CTX_SYNC_MEMOPS flag for the current context */
+    uct_cuda_copy_sync_memops(address, is_vmm);
+
+    if (md->config.alloc_whole_reg == UCS_CONFIG_OFF) {
+        /* Extending the registration range is disable by configuration */
+        goto out_ctx_pop;
+    }
+
+    if (UCT_CUDADRV_FUNC_LOG_DEBUG(
+                cuMemGetAddressRange(&base_address, &alloc_length, address)) !=
+        UCS_OK) {
+        goto out_ctx_pop;
+    }
+
+    ucs_trace("query address 0x%llx: 0x%llx..0x%llx length %zu", address,
+              base_address, base_address + alloc_length, alloc_length);
+
+    if (md->config.alloc_whole_reg == UCS_CONFIG_AUTO) {
+        total_bytes = uct_cuda_copy_md_get_total_device_mem(cuda_device);
+        if (alloc_length > (total_bytes * md->config.max_reg_ratio)) {
+            goto out_ctx_pop;
+        }
+    } else {
+        ucs_assert(md->config.alloc_whole_reg == UCS_CONFIG_ON);
+    }
+
+    mem_info->base_address = (void*)base_address;
+    mem_info->alloc_length = alloc_length;
+
+out_ctx_pop:
+    UCT_CUDADRV_FUNC_LOG_WARN(cuCtxPopCurrent(&tmp_ctx));
+    if (cuda_ctx == NULL) {
+        UCT_CUDADRV_FUNC_LOG_WARN(cuDevicePrimaryCtxRelease(cuda_device));
+    }
+}
+
 static ucs_status_t
 uct_cuda_copy_md_query_attributes(uct_cuda_copy_md_t *md, const void *address,
                                   size_t length, ucs_memory_info_t *mem_info)
@@ -582,11 +748,8 @@ uct_cuda_copy_md_query_attributes(uct_cuda_copy_md_t *md, const void *address,
     CUcontext cuda_mem_ctx     = NULL;
     CUpointer_attribute attr_type[UCT_CUDA_MEM_QUERY_NUM_ATTRS];
     void *attr_data[UCT_CUDA_MEM_QUERY_NUM_ATTRS];
-    CUdeviceptr base_address;
-    size_t alloc_length;
-    size_t total_bytes;
     int32_t pref_loc;
-    unsigned is_vmm;
+    int is_vmm;
     CUresult cu_err;
     ucs_status_t status;
 
@@ -620,21 +783,10 @@ uct_cuda_copy_md_query_attributes(uct_cuda_copy_md_t *md, const void *address,
             return UCS_ERR_INVALID_ADDR;
         }
 
-        if (is_managed ||
-            ((cuda_mem_ctx == NULL) && md->config.cuda_async_managed)) {
-            /* is_managed: cuMemGetAddress range does not support managed memory
-             * so use provided address and length as base address and alloc
-             * length respectively
-             *
-             * cuda_async_managed: currently virtual/stream-ordered CUDA
-             * allocations are typed as `UCS_MEMORY_TYPE_CUDA_MANAGED`. This may
-             * be changed using UCX_CUDA_COPY_ASYNC_MEM_TYPE env var.
-             * Ideally checking for
-             * `CU_POINTER_ATTRIBUTE_IS_LEGACY_CUDA_IPC_CAPABLE` would be better
-             * here, but due to a bug in the driver `cudaMalloc` also returns false
-             * in that case. Therefore, checking whether the allocation was not
-             * allocated in a context should also allows us to identify
-             * virtual/stream-ordered CUDA allocations. */
+        if (is_managed) {
+            /* cuMemGetAddress range does not support managed memory so use
+             * provided address and length as base address and alloc length
+             * respectively */
             mem_info->type = UCS_MEMORY_TYPE_CUDA_MANAGED;
 
             cu_err = cuMemRangeGetAttribute(
@@ -659,6 +811,16 @@ uct_cuda_copy_md_query_attributes(uct_cuda_copy_md_t *md, const void *address,
             }
 
             goto out_default_range;
+        } else if ((cuda_mem_ctx == NULL) && md->config.cuda_async_managed) {
+            /* Currently virtual/stream-ordered CUDA allocations are typed as
+             * `UCS_MEMORY_TYPE_CUDA_MANAGED`. This may be changed using
+             * UCX_CUDA_COPY_ASYNC_MEM_TYPE env var. Ideally checking for
+             * `CU_POINTER_ATTRIBUTE_IS_LEGACY_CUDA_IPC_CAPABLE` would be better
+             * here, but due to a bug in the driver `cudaMalloc` also returns
+             * false in that case. Therefore, checking whether the allocation
+             * was not allocated in a context should also allows us to
+             * identify virtual/stream-ordered CUDA allocations. */
+            mem_info->type = UCS_MEMORY_TYPE_CUDA_MANAGED;
         } else {
             mem_info->type = UCS_MEMORY_TYPE_CUDA;
         }
@@ -674,35 +836,10 @@ uct_cuda_copy_md_query_attributes(uct_cuda_copy_md_t *md, const void *address,
         return UCS_ERR_NO_DEVICE;
     }
 
-    uct_cuda_copy_sync_memops(md, address, is_vmm);
-
-    /* Extending the registration range is disable by configuration */
-    if (md->config.alloc_whole_reg == UCS_CONFIG_OFF) {
-        goto out_default_range;
-    }
-
-    cu_err = cuMemGetAddressRange(&base_address, &alloc_length,
-                                  (CUdeviceptr)address);
-    if (cu_err != CUDA_SUCCESS) {
-        ucs_error("cuMemGetAddressRange(%p) error: %s", address,
-                  uct_cuda_base_cu_get_error_string(cu_err));
-        return UCS_ERR_INVALID_ADDR;
-    }
-
-    ucs_trace("query address %p: 0x%llx..0x%llx length %zu", address,
-              base_address, base_address + alloc_length, alloc_length);
-
-    if (md->config.alloc_whole_reg == UCS_CONFIG_AUTO) {
-        total_bytes = uct_cuda_copy_md_get_total_device_mem(cuda_device);
-        if (alloc_length > (total_bytes * md->config.max_reg_ratio)) {
-            goto out_default_range;
-        }
-    } else {
-        ucs_assert(md->config.alloc_whole_reg == UCS_CONFIG_ON);
-    }
-
-    mem_info->base_address = (void*)base_address;
-    mem_info->alloc_length = alloc_length;
+    uct_cuda_copy_md_sync_memops_get_address_range(md, (CUdeviceptr)address,
+                                                   length, cuda_mem_ctx,
+                                                   cuda_device, is_vmm,
+                                                   mem_info);
     return UCS_OK;
 
 out_default_range:
@@ -714,7 +851,7 @@ out_default_range:
 static int uct_cuda_copy_md_get_dmabuf_fd(uintptr_t address, size_t length)
 {
 #if CUDA_VERSION >= 11070
-    PFN_cuMemGetHandleForAddressRange get_handle_func;
+    PFN_cuMemGetHandleForAddressRange_v11070 get_handle_func;
     CUresult cu_err;
     int fd;
 
@@ -756,7 +893,7 @@ static int uct_cuda_copy_md_get_dmabuf_fd(uintptr_t address, size_t length)
     return UCT_DMABUF_FD_INVALID;
 }
 
-static ucs_status_t
+ucs_status_t
 uct_cuda_copy_md_mem_query(uct_md_h tl_md, const void *address, size_t length,
                            uct_md_mem_attr_t *mem_attr)
 {
@@ -853,11 +990,12 @@ static uct_md_ops_t md_ops = {
     .query              = uct_cuda_copy_md_query,
     .mem_alloc          = uct_cuda_copy_mem_alloc,
     .mem_free           = uct_cuda_copy_mem_free,
-    .mkey_pack          = uct_cuda_copy_mkey_pack,
+    .mem_advise         = (uct_md_mem_advise_func_t)ucs_empty_function_return_unsupported,
     .mem_reg            = uct_cuda_copy_mem_reg,
     .mem_dereg          = uct_cuda_copy_mem_dereg,
-    .mem_attach         = ucs_empty_function_return_unsupported,
     .mem_query          = uct_cuda_copy_md_mem_query,
+    .mkey_pack          = (uct_md_mkey_pack_func_t)ucs_empty_function_return_success,
+    .mem_attach         = (uct_md_mem_attach_func_t)ucs_empty_function_return_unsupported,
     .detect_memory_type = uct_cuda_copy_md_detect_memory_type
 };
 
@@ -878,15 +1016,15 @@ uct_cuda_copy_md_open(uct_component_t *component, const char *md_name,
         goto err;
     }
 
-    md->super.ops               = &md_ops;
-    md->super.component         = &uct_cuda_copy_component;
-    md->config.alloc_whole_reg  = config->alloc_whole_reg;
-    md->config.max_reg_ratio    = config->max_reg_ratio;
-    md->config.pref_loc         = config->pref_loc;
-    md->config.enable_fabric    = config->enable_fabric;
-    md->config.dmabuf_supported = 0;
-    md->sync_memops_set         = 0;
-    md->granularity             = SIZE_MAX;
+    md->super.ops                 = &md_ops;
+    md->super.component           = &uct_cuda_copy_component;
+    md->config.alloc_whole_reg    = config->alloc_whole_reg;
+    md->config.max_reg_ratio      = config->max_reg_ratio;
+    md->config.pref_loc           = config->pref_loc;
+    md->config.enable_fabric      = config->enable_fabric;
+    md->config.dmabuf_supported   = 0;
+    md->config.retain_primary_ctx = config->retain_primary_ctx;
+    md->granularity               = SIZE_MAX;
 
     if ((config->cuda_async_mem_type != UCS_MEMORY_TYPE_CUDA) &&
         (config->cuda_async_mem_type != UCS_MEMORY_TYPE_CUDA_MANAGED)) {
@@ -911,6 +1049,13 @@ uct_cuda_copy_md_open(uct_component_t *component, const char *md_name,
 
     *md_p = (uct_md_h)md;
 
+    /*
+     * Setting sync memops flag for the first time during memory detection can
+     * cause a deadlock if other CUDA operations are performed in parallel.
+     * We avoid that issue by preemptively setting it during MD open.
+     */
+    uct_cuda_copy_set_ctx_sync_memops(UCS_LOG_LEVEL_DEBUG);
+
     return UCS_OK;
 
 err_free_md:
@@ -922,10 +1067,10 @@ err:
 uct_component_t uct_cuda_copy_component = {
     .query_md_resources = uct_cuda_base_query_md_resources,
     .md_open            = uct_cuda_copy_md_open,
-    .cm_open            = ucs_empty_function_return_unsupported,
-    .rkey_unpack        = uct_cuda_copy_rkey_unpack,
-    .rkey_ptr           = ucs_empty_function_return_unsupported,
-    .rkey_release       = uct_cuda_copy_rkey_release,
+    .cm_open            = (uct_component_cm_open_func_t)ucs_empty_function_return_unsupported,
+    .rkey_unpack        = uct_md_stub_rkey_unpack,
+    .rkey_ptr           = (uct_component_rkey_ptr_func_t)ucs_empty_function_return_unsupported,
+    .rkey_release       = (uct_component_rkey_release_func_t)ucs_empty_function_return_success,
     .rkey_compare       = uct_base_rkey_compare,
     .name               = "cuda_cpy",
     .md_config          = {
