@@ -14,12 +14,13 @@
 #include <ucs/config/types.h>
 
 
-#if HAVE_CUDA_FABRIC
 typedef enum uct_cuda_ipc_key_handle {
-    UCT_CUDA_IPC_KEY_HANDLE_TYPE_ERROR = 0,
+    UCT_CUDA_IPC_KEY_HANDLE_TYPE_NO_IPC = 0,
     UCT_CUDA_IPC_KEY_HANDLE_TYPE_LEGACY, /* cudaMalloc memory */
+#if HAVE_CUDA_FABRIC
     UCT_CUDA_IPC_KEY_HANDLE_TYPE_VMM, /* cuMemCreate memory */
     UCT_CUDA_IPC_KEY_HANDLE_TYPE_MEMPOOL /* cudaMallocAsync memory */
+#endif
 } uct_cuda_ipc_key_handle_t;
 
 
@@ -27,15 +28,16 @@ typedef struct uct_cuda_ipc_md_handle {
     uct_cuda_ipc_key_handle_t handle_type;
     union {
         CUipcMemHandle        legacy;        /* Legacy IPC handle */
+#if HAVE_CUDA_FABRIC
         CUmemFabricHandle     fabric_handle; /* VMM/Mallocasync export handle */
+#endif
     } handle;
+#if HAVE_CUDA_FABRIC
     CUmemPoolPtrExportData    ptr;
     CUmemoryPool              pool;
-} uct_cuda_ipc_md_handle_t;
-#else
-typedef CUipcMemHandle uct_cuda_ipc_md_handle_t;
 #endif
-
+    unsigned long long        buffer_id;
+} uct_cuda_ipc_md_handle_t;
 
 /**
  * @brief cuda ipc MD descriptor
@@ -133,11 +135,16 @@ typedef struct {
     pid_t                     pid;     /* PID as key to resolve peer_map hash */
     CUdeviceptr               d_bptr;  /* Allocation base address */
     size_t                    b_len;   /* Allocation size */
-    int                       dev_num; /* GPU Device number */
     CUuuid                    uuid;    /* GPU Device UUID */
 } uct_cuda_ipc_rkey_t;
 
 
+typedef struct {
+    uct_cuda_ipc_rkey_t       super;
+    int                       stream_id;
+} uct_cuda_ipc_unpacked_rkey_t;
+
+
 #define UCT_CUDA_IPC_GET_DEVICE(_cu_device)                          \
     do {                                                             \
         if (UCS_OK !=                                                \
