@@ -23,6 +23,16 @@ uct_dc_mlx5_add_flush_remote(uct_dc_mlx5_ep_t *ep)
     ep->flags |= UCT_DC_MLX5_EP_FLAG_FLUSH_REMOTE;
 }
 
+static UCS_F_ALWAYS_INLINE uint16_t
+uct_dc_mlx5_atomic_offset(uct_dc_mlx5_ep_t *ep)
+{
+    if (ep->flags & UCT_DC_MLX5_EP_FLAG_ATOMIC_OFFSET) {
+        return uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    }
+
+    return 0;
+}
+
 static UCS_F_ALWAYS_INLINE size_t
 uct_dc_mlx5_set_dgram_seg(uct_ib_mlx5_txwq_t *txwq,
                           uct_dc_mlx5_iface_t *iface,
@@ -109,7 +119,7 @@ uct_dc_mlx5_iface_atomic_post(uct_dc_mlx5_iface_t *iface, uct_dc_mlx5_ep_t *ep,
                               uint64_t compare_mask, uint64_t compare,
                               uint64_t swap_mask, uint64_t swap_add)
 {
-    uint16_t atomic_mr_offset = uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    uint16_t atomic_mr_offset = uct_dc_mlx5_atomic_offset(ep);
     uint32_t ib_rkey          = uct_ib_resolve_atomic_rkey(rkey,
                                                            atomic_mr_offset,
                                                            &remote_addr);
@@ -263,10 +273,17 @@ ucs_status_t uct_dc_mlx5_ep_fence(uct_ep_h tl_ep, unsigned flags)
 {
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(tl_ep->iface, uct_dc_mlx5_iface_t);
     uct_dc_mlx5_ep_t *ep       = ucs_derived_of(tl_ep, uct_dc_mlx5_ep_t);
+    uct_dc_dci_t *dci;
 
-    return uct_rc_ep_fence(tl_ep,
-                           &uct_dc_mlx5_iface_dci(iface, ep->dci)->txwq.fi,
-                           ep->dci != UCT_DC_MLX5_EP_NO_DCI);
+    if (ep->dci != UCT_DC_MLX5_EP_NO_DCI) {
+        dci = uct_dc_mlx5_iface_dci(iface, ep->dci);
+        ucs_assertv(uct_dc_mlx5_is_dci_valid(dci), "iface=%p ep=%p dci=%d",
+                    iface, ep, ep->dci);
+        return uct_rc_ep_fence(tl_ep, &dci->txwq.fi);
+    }
+
+    UCT_TL_EP_STAT_FENCE(&ep->super);
+    return UCS_OK;
 }
 
 static ucs_status_t UCS_F_ALWAYS_INLINE
@@ -465,7 +482,7 @@ uct_dc_mlx5_ep_put_short_inline(uct_ep_h tl_ep, const void *buffer,
 {
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(tl_ep->iface, uct_dc_mlx5_iface_t);
     uct_dc_mlx5_ep_t *ep       = ucs_derived_of(tl_ep, uct_dc_mlx5_ep_t);
-    uint16_t atomic_mr_offset  = uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    uint16_t atomic_mr_offset  = uct_dc_mlx5_atomic_offset(ep);
     size_t av_size;
     UCT_DC_MLX5_TXQP_DECL(txqp, txwq);
 
@@ -495,7 +512,7 @@ ucs_status_t uct_dc_mlx5_ep_put_short(uct_ep_h tl_ep, const void *payload,
 #if HAVE_IBV_DM
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(tl_ep->iface, uct_dc_mlx5_iface_t);
     uct_dc_mlx5_ep_t *ep       = ucs_derived_of(tl_ep, uct_dc_mlx5_ep_t);
-    uint16_t atomic_mr_offset  = uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    uint16_t atomic_mr_offset  = uct_dc_mlx5_atomic_offset(ep);
     ucs_status_t status;
     size_t av_size;
     UCT_DC_MLX5_TXQP_DECL(txqp, txwq);
@@ -535,7 +552,7 @@ ssize_t uct_dc_mlx5_ep_put_bcopy(uct_ep_h tl_ep, uct_pack_callback_t pack_cb,
 {
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(tl_ep->iface, uct_dc_mlx5_iface_t);
     uct_dc_mlx5_ep_t *ep       = ucs_derived_of(tl_ep, uct_dc_mlx5_ep_t);
-    uint16_t atomic_mr_offset  = uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    uint16_t atomic_mr_offset  = uct_dc_mlx5_atomic_offset(ep);
     uct_rc_iface_send_desc_t *desc;
     size_t length;
     UCT_DC_MLX5_TXQP_DECL(txqp, txwq);
@@ -559,7 +576,7 @@ ucs_status_t uct_dc_mlx5_ep_put_zcopy(uct_ep_h tl_ep, const uct_iov_t *iov, size
 {
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(tl_ep->iface, uct_dc_mlx5_iface_t);
     uct_dc_mlx5_ep_t *ep       = ucs_derived_of(tl_ep, uct_dc_mlx5_ep_t);
-    uint16_t atomic_mr_offset  = uct_ib_md_atomic_offset(ep->atomic_mr_id);
+    uint16_t atomic_mr_offset  = uct_dc_mlx5_atomic_offset(ep);
     UCT_DC_MLX5_TXQP_DECL(txqp, txwq);
 
     UCT_CHECK_IOV_SIZE(iovcnt, UCT_RC_MLX5_RMA_MAX_IOV(UCT_IB_MLX5_AV_FULL_SIZE),
@@ -1248,6 +1265,10 @@ UCS_CLASS_INIT_FUNC(uct_dc_mlx5_ep_t, uct_dc_mlx5_iface_t *iface,
         self->flush_rkey_hi = 0;
     }
 
+    if (!(if_addr->flags & UCT_DC_MLX5_IFACE_ADDR_NO_ATOMIC_OFFSET)) {
+        self->flags |= UCT_DC_MLX5_EP_FLAG_ATOMIC_OFFSET;
+    }
+
     if ((iface->super.super.config.max_rd_atomic == 16) !=
         !!(if_addr->flags & UCT_DC_MLX5_IFACE_ADDR_MAX_RD_ATOMIC_16)) {
         ucs_diag("max_rd_atomic values do not match on peers (local is %u), "
@@ -1276,7 +1297,7 @@ UCS_CLASS_CLEANUP_FUNC(uct_dc_mlx5_ep_t)
 {
     uct_dc_mlx5_iface_t *iface = ucs_derived_of(self->super.super.iface,
                                                 uct_dc_mlx5_iface_t);
-    uct_dc_dci_t *dci          = uct_dc_mlx5_iface_dci(iface, self->dci);
+    uct_dc_dci_t *dci;
 
     uct_dc_mlx5_ep_pending_purge(&self->super.super,
                                  uct_rc_ep_pending_purge_warn_cb, self);
@@ -1294,7 +1315,7 @@ UCS_CLASS_CLEANUP_FUNC(uct_dc_mlx5_ep_t)
     ucs_assertv_always(uct_dc_mlx5_iface_dci_has_outstanding(iface, self->dci),
                        "iface (%p) ep (%p) dci leak detected: dci=%d", iface,
                        self, self->dci);
-
+    dci = uct_dc_mlx5_iface_dci(iface, self->dci);
     if (!uct_dc_mlx5_is_dci_valid(dci)) {
         return;
     }
@@ -1401,7 +1422,12 @@ uct_dc_mlx5_iface_dci_do_pending_wait(ucs_arbiter_t *arbiter,
     }
 
     ucs_assert_always(ep->dci != UCT_DC_MLX5_EP_NO_DCI);
-    uct_dc_mlx5_iface_dci_sched_tx(iface, ep);
+
+    if (uct_dc_mlx5_iface_dci_has_tx_resources(iface, ep->dci)) {
+        ucs_arbiter_group_schedule(uct_dc_mlx5_iface_tx_waitq(iface),
+                                   &ep->arb_group);
+    }
+
     return UCS_ARBITER_CB_RESULT_DESCHED_GROUP;
 }
 
@@ -1460,7 +1486,11 @@ unsigned uct_dc_mlx5_ep_dci_release_progress(void *arg)
         dci_pool = &iface->tx.dci_pool[pool_index];
         while (dci_pool->release_stack_top >= 0) {
             dci = ucs_array_elem(&dci_pool->stack, dci_pool->release_stack_top--);
-            ucs_assert(dci < iface->tx.ndci * iface->tx.num_dci_pools);
+            ucs_assertv((dci - uct_dc_mlx5_iface_is_hybrid(iface)) <
+                                (iface->tx.ndci * iface->tx.num_dci_pools),
+                        "dci=%d is_hybrid=%s, ndci=%d, num_dci_pools=%d", dci,
+                        uct_dc_mlx5_iface_is_hybrid(iface) ? "yes" : "no",
+                        iface->tx.ndci, iface->tx.num_dci_pools);
             uct_dc_mlx5_iface_dci_release(iface, dci);
         }
 
@@ -1476,6 +1506,21 @@ unsigned uct_dc_mlx5_ep_dci_release_progress(void *arg)
     return 1;
 }
 
+static ucs_arbiter_cb_result_t
+uct_dc_mlx5_iface_shared_dci_pending_resched(ucs_arbiter_cb_result_t res,
+                                             uct_dc_mlx5_iface_t *iface,
+                                             uct_dc_mlx5_ep_t *ep)
+{
+    if ((res == UCS_ARBITER_CB_RESULT_DESCHED_GROUP) &&
+        uct_rc_fc_has_resources(&iface->super.super, &ep->fc)) {
+        /* We can't desched group with a shared dci if non FC resources are
+         * missing, since it's never scheduled again. */
+        res = UCS_ARBITER_CB_RESULT_RESCHED_GROUP;
+    }
+
+    return res;
+}
+
 /**
  * dispatch requests waiting for tx resources (dcs* DCI policies)
  */
@@ -1494,6 +1539,11 @@ uct_dc_mlx5_iface_dci_do_dcs_pending_tx(ucs_arbiter_t *arbiter,
     ucs_arbiter_cb_result_t res;
 
     res = uct_dc_mlx5_iface_dci_do_common_pending_tx(ep, elem);
+
+    if (uct_dc_mlx5_is_hw_dci(iface, ep->dci)) {
+        return uct_dc_mlx5_iface_shared_dci_pending_resched(res, iface, ep);
+    }
+
     if ((res != UCS_ARBITER_CB_RESULT_REMOVE_ELEM) || !is_only) {
         return res;
     }
@@ -1522,14 +1572,7 @@ uct_dc_mlx5_iface_dci_do_rand_pending_tx(ucs_arbiter_t *arbiter,
     ucs_arbiter_cb_result_t res;
 
     res = uct_dc_mlx5_iface_dci_do_common_pending_tx(ep, elem);
-    if ((res == UCS_ARBITER_CB_RESULT_DESCHED_GROUP) &&
-        uct_rc_fc_has_resources(&iface->super.super, &ep->fc)) {
-        /* We can't desched group with rand policy if non FC resources are
-         * missing, since it's never scheduled again. */
-        res = UCS_ARBITER_CB_RESULT_RESCHED_GROUP;
-    }
-
-    return res;
+    return uct_dc_mlx5_iface_shared_dci_pending_resched(res, iface, ep);
 }
 
 static ucs_arbiter_cb_result_t
@@ -1710,23 +1753,22 @@ void uct_dc_mlx5_ep_handle_failure(uct_dc_mlx5_ep_t *ep,
                                                 uct_dc_mlx5_iface_t);
     uct_dci_index_t dci_index  = ep->dci;
     uint16_t pi                = ntohs(cqe->wqe_counter);
+    uct_dc_dci_t *dci          = uct_dc_mlx5_iface_dci(iface, dci_index);
     uint8_t pool_index;
-    UCT_DC_MLX5_TXQP_DECL(txqp, txwq);
-
-    UCT_DC_MLX5_IFACE_TXQP_DCI_GET(iface, dci_index, txqp, txwq);
 
     ucs_debug("handle failure iface %p, ep %p, dci[%d] qpn 0x%x, status: %s",
-              iface, ep, dci_index, txwq->super.qp_num,
+              iface, ep, dci_index, dci->txwq.super.qp_num,
               ucs_status_string(ep_status));
 
     ucs_assert(!uct_dc_mlx5_iface_is_policy_shared(iface));
 
-    uct_dc_mlx5_update_tx_res(iface, txwq, txqp, pi);
-    uct_rc_txqp_purge_outstanding(&iface->super.super, txqp, ep_status, pi, 0);
+    uct_dc_mlx5_update_tx_res(iface, dci_index, pi);
+    uct_rc_txqp_purge_outstanding(&iface->super.super, &dci->txqp, ep_status,
+                                  pi, 0);
 
     /* Invoke a user's error callback and release TX/FC resources before
      * releasing DCI, to have DCI for doing possible flush(CANCEL) */
-    uct_dc_mlx5_iface_set_ep_failed(iface, ep, cqe, txwq, ep_status);
+    uct_dc_mlx5_iface_set_ep_failed(iface, ep, cqe, &dci->txwq, ep_status);
 
     ucs_assertv(ep->fc.fc_wnd == iface->super.super.config.fc_wnd_size,
                 "ep %p: fc_wnd=%d config_fc_wnd=%u", ep, ep->fc.fc_wnd,
