@@ -72,6 +72,7 @@ enum {
     UCT_IB_MLX5_CMD_OP_2ERR_QP                 = 0x507,
     UCT_IB_MLX5_CMD_OP_2RST_QP                 = 0x50a,
     UCT_IB_MLX5_CMD_OP_QUERY_QP                = 0x50b,
+    UCT_IB_MLX5_CMD_OP_INIT2INIT_QP            = 0x50e,
     UCT_IB_MLX5_CMD_OP_CREATE_RMP              = 0x90c,
     UCT_IB_MLX5_CMD_OP_CREATE_DCT              = 0x710,
     UCT_IB_MLX5_CMD_OP_DRAIN_DCT               = 0x712,
@@ -201,7 +202,10 @@ struct uct_ib_mlx5_cmd_hca_cap_bits {
     uint8_t    pps[0x1];
     uint8_t    pps_modify[0x1];
     uint8_t    log_max_msg[0x5];
-    uint8_t    reserved_at_1c8[0x4];
+    uint8_t    dp_ordering_ooo_rw_xrc[0x1];
+    uint8_t    dp_ordering_ooo_rw_dc[0x1];
+    uint8_t    dp_ordering_ooo_rw_rc[0x1];
+    uint8_t    reserved_at_1cb[0x1];
     uint8_t    max_tc[0x4];
     uint8_t    reserved_at_1d0[0x1];
     uint8_t    dcbx[0x1];
@@ -319,7 +323,12 @@ struct uct_ib_mlx5_cmd_hca_cap_bits {
     uint8_t    log_max_transport_domain[0x5];
     uint8_t    reserved_at_328[0x3];
     uint8_t    log_max_pd[0x5];
-    uint8_t    reserved_at_330[0xb];
+    uint8_t    dp_ordering_ooo_all_ud[0x1];
+    uint8_t    dp_ordering_ooo_all_uc[0x1];
+    uint8_t    dp_ordering_ooo_all_xrc[0x1];
+    uint8_t    dp_ordering_ooo_all_dc[0x1];
+    uint8_t    dp_ordering_ooo_all_rc[0x1];
+    uint8_t    reserved_at_335[0x6];
     uint8_t    log_max_xrcd[0x5];
 
     uint8_t    nic_receive_steering_discard[0x1];
@@ -428,7 +437,9 @@ struct uct_ib_mlx5_cmd_hca_cap_bits {
     uint8_t    affiliate_nic_vport_criteria[0x8];
     uint8_t    native_port_num[0x8];
     uint8_t    num_vhca_ports[0x8];
-    uint8_t    reserved_at_618[0x6];
+    uint8_t    flex_parser_id_gtpu_teid[0x4];
+    uint8_t    reserved_at_61c[0x1];
+    uint8_t    trusted_vnic_vhca[0x1];
     uint8_t    sw_owner_id[0x1];
     uint8_t    reserve_not_to_use[0x1];
     uint8_t    reserved_at_620[0x60];
@@ -497,7 +508,8 @@ struct uct_ib_mlx5_cmd_hca_cap_2_bits {
 
     uint8_t    reserved_at_80[0x3];
     uint8_t    max_num_prog_sample_field[0x5];
-    uint8_t    reserved_at_88[0x3];
+    uint8_t    dp_ordering_force[0x1];
+    uint8_t    reserved_at_89[0x2];
     uint8_t    log_max_num_reserved_qpn[0x5];
     uint8_t    atomic_rate_pa[0x1];
     uint8_t    introspection_mkey_access_allowed[0x1];
@@ -867,7 +879,7 @@ struct uct_ib_mlx5_set_xrq_dc_params_entry_in_bits {
     uint8_t         reserved_at_80[0x3];
     uint8_t         ack_timeout[0x5];
     uint8_t         reserved_at_88[0x4];
-    uint8_t         multi_path[0x1];
+    uint8_t         dp_ordering[0x1];
     uint8_t         mtu[0x3];
     uint8_t         pkey_table_index[0x10];
 
@@ -910,7 +922,8 @@ struct uct_ib_mlx5_dctc_bits {
     uint8_t         offload_type[0x4];
     uint8_t         reserved_at_1c[0x4];
 
-    uint8_t         reserved_at_20[0x8];
+    uint8_t         reserved_at_20[0x7];
+    uint8_t         dp_ordering_force[0x1];
     uint8_t         user_index[0x18];
 
     uint8_t         reserved_at_40[0x8];
@@ -925,8 +938,9 @@ struct uct_ib_mlx5_dctc_bits {
     uint8_t         latency_sensitive[0x1];
     uint8_t         rlky[0x1];
     uint8_t         force_full_handshake[0x1];
-    uint8_t         multi_path[0x1];
-    uint8_t         reserved_at_74[0xc];
+    uint8_t         dp_ordering_0[0x1];
+    uint8_t         dp_ordering_1[0x1];
+    uint8_t         reserved_at_75[0xb];
 
     uint8_t         reserved_at_80[0x8];
     uint8_t         cs_res[0x8];
@@ -1295,6 +1309,10 @@ enum {
     UCT_IB_MLX5_QPC_OFFLOAD_TYPE_RNDV  = 0x1
 };
 
+enum {
+    UCT_IB_MLX5_QPC_OPT_MASK_32_INIT2INIT_MMO = 1 << 3,
+};
+
 enum {
     UCT_IB_MLX5_QPC_END_PADDING_MODE_SCATTER_AS_IS                = 0x0,
     UCT_IB_MLX5_QPC_END_PADDING_MODE_PAD_TO_CACHE_LINE_ALIGNMENT  = 0x1
@@ -1377,7 +1395,8 @@ struct uct_ib_mlx5_qpc_bits {
     uint8_t         latency_sensitive[0x1];
     uint8_t         reserved_at_24[0x1];
     uint8_t         drain_sigerr[0x1];
-    uint8_t         reserved_at_26[0x2];
+    uint8_t         dp_ordering_0[0x1];
+    uint8_t         dp_ordering_force[0x1];
     uint8_t         pd[0x18];
 
     uint8_t         mtu[0x3];
@@ -1451,7 +1470,8 @@ struct uct_ib_mlx5_qpc_bits {
     uint8_t         rae[0x1];
     uint8_t         reserved_at_493[0x1];
     uint8_t         page_offset[0x6];
-    uint8_t         reserved_at_49a[0x3];
+    uint8_t         reserved_at_49a[0x2];
+    uint8_t         dp_ordering_1[0x1];
     uint8_t         cd_slave_receive[0x1];
     uint8_t         cd_slave_send[0x1];
     uint8_t         cd_master[0x1];
@@ -1502,6 +1522,20 @@ struct uct_ib_mlx5_qpc_bits {
     uint8_t         dbr_umem_id[0x20];
 };
 
+struct uct_ib_mlx5_qpc_ext_bits {
+    uint8_t         delay_drop_en[0x1];
+    uint8_t         vl15[0x1];
+    uint8_t         mmo[0x1];
+    uint8_t         reserved_at_3[0xd];
+    uint8_t         dci_stream_channel_id[0x10];
+
+    uint8_t         qos_queue_group_id_requester[0x20];
+
+    uint8_t         qos_queue_group_id_responder[0x20];
+
+    uint8_t         reserved_at_80[0x5a0];
+};
+
 struct uct_ib_mlx5_create_qp_out_bits {
     uint8_t         status[0x8];
     uint8_t         reserved_at_8[0x18];
@@ -1539,6 +1573,42 @@ struct uct_ib_mlx5_create_qp_in_bits {
     uint8_t         pas[0][0x40];
 };
 
+struct uct_ib_mlx5_init2init_qp_out_bits {
+    uint8_t         status[0x8];
+    uint8_t         reserved_at_8[0x18];
+
+    uint8_t         syndrome[0x20];
+
+    uint8_t         reserved_at_40[0x20];
+    uint8_t         ece[0x20];
+};
+
+struct uct_ib_mlx5_init2init_qp_in_bits {
+    uint8_t                         opcode[0x10];
+    uint8_t                         uid[0x10];
+
+    uint8_t                         vhca_tunnel_id[0x10];
+    uint8_t                         op_mod[0x10];
+
+    uint8_t                         qpc_ext[0x1];
+    uint8_t                         reserved_at_41[0x7];
+    uint8_t                         qpn[0x18];
+
+    uint8_t                         reserved_at_60[0x20];
+
+    uint8_t                         opt_param_mask[0x20];
+
+    uint8_t                         ece[0x20];
+
+    struct uct_ib_mlx5_qpc_bits     qpc_data;
+
+    uint8_t                         reserved_at_800[0x40];
+
+    uint8_t                         opt_param_mask_95_32[0x40];
+
+    struct uct_ib_mlx5_qpc_ext_bits qpc_data_extension;
+};
+
 struct uct_ib_mlx5_init2rtr_qp_out_bits {
     uint8_t         status[0x8];
     uint8_t         reserved_at_8[0x18];
